/* mscalc.cc			letzte Aenderung: 11.11.2014 */
#define VERSION "Version 0.3"
/*
Uebersetzen auf Unix (Linux):
> make  ;siehe makefile

 Kurzbeschreibung: Massenspektrum (MS) Isotopenaufspaltung berechnen

History:
11.11.2014      Erstellung aus mssim.cc (RP)

*/

#include <stdio.h>
//#include <iostream>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <xtekplot1.h>

#define SUMMENFORMELN
#include "mgkern.c"

#define XMAX 1024
#define YMAX  820
//#define TIEFE 24
#define TIEFE 8

#define SCHWARZAUFWEISS //auskommentieren fuer weiss auf schwarz

/************************* Vordeklarationen ***************************/
double mg(const char *formel);

/**************** Routinen zur Parameterauswertung ********************/
#define MAXARG 2
static char argflag[128];
void setargflags(char *s)
{
 int c;
 while(c= *s++)
  {if(c>='a' && c<='z')  c -= 'a'-'A';
   argflag[c&127]=1;
  }
}
/*************************** kleinkram ***************************/

/*********************** Zufallszahlen ***************************/
double zufall4()	/* Zufallszahl zwischen 0.0 und 1.0 */
{			/* bisher bester Zufallsgenerator   */
 static double x=581;
 double y;
 if(x==0.) x=581;
 x/=1.163;
 x-=(long)(x);
 y=10.*x;
 x*=1000.;
 return (y-(long)(y));
}

#include <time.h>
double zufall5(int start=0)
{
 static int flag=1;
 if(flag!=0 || start!=0)
  {time_t t;
   time(&t);
   srand(t);
   flag=0;
  }
 return (1.0 - rand()/(double)RAND_MAX);
}

/******************** Wahrscheinlichkeitsrechnung *********************/
struct elements {const char *sym; double m1,m2, w1,w2;};
elements element[]=
{
 "C",	12.0000,13.00335, 0.9893,0.0107,  //Durchschnitt: 12.0107
 "Cu",	62.9296,64.9278, 0.6915,0.3085,
 "N",   14.0003,15.0001, 0.99636,0.00364,
 "O",   15.9949,17.9992, 0.99757,0.00205,  //noch nicht beruecksichtigt: ^17O 16.9991 0.00038
 "S",   31.9721,33.9679, 0.9499,0.0425,    //^33S 32.9715 0.0075, ^36S 35.9671 0.0001
 "H",   1.00794,2.0141, 0.999885,0.000115,
 "Cl",  34.9688,36.9659, 0.7576,0.2424,
 "Br",  78.9183,80.9163, 0.5069,0.4931,
 "B",   11.0093,10.0129, 0.801,0.199,
 "Li",   7.0160,6.01512, 0.9241,0.0759,
 "Mg",  23.9850,25.9826, 0.7899,0.1101,  //^25Mg 24.9858 0.1000
 "Si",  27.9769,28.9765, 0.92223,0.04685, //^30Si 29.9738 0.03092
 "K",   38.9637,40.9618, 0.93258,0.067302,
 "Ca",  39.9626,43.9555, 0.96941,0.02086, //diverse weitere Isotope noch nicht beruecksichtigt
 "Fe",  55.9349,53.9396, 0.91754,0.05845,
 "Ni",  57.9353,59.9308, 0.680769,0.262231,
 "Zn",  63.9291,65.9260, 0.48268,0.27975, //^68Zn 67.9248 0.19024
 "Se",  79.9165,77.9173, 0.4961,0.2377, //diverse weitere Isotope noch nicht beruecksichtigt
 "Ag", 106.9051,108.9047, 0.51839,0.48161,
 "Ti",  47.9479, 45.9526, 0.7372,0.0825, //diverse weitere Isotope noch nicht beruecksichtigt
 "V",   50.9440, 49.9472, 0.9975,0.0025,
 "Cr",  51.9405, 52.9406, 0.83789,0.09501,
 "Ga",  68.9256, 70.9247, 0.60108,0.39892,
 "Ge",  73.9212, 71.9221, 0.3672,0.2731,
 //bisher alle Elemente bis Brom, ausser den Edelgasen, beruecksichtigt
 //aber bisher nur jeweils 2 Isotope pro Element beruecksichtigt.
 NULL,	0.,0.,		0.,0
};

struct wahrtab {double a,b,m1,m2;};
wahrtab wahr['z'*256+'Z'];
void init_wahr(void)
{
 elements *p;
 int index;
 for(p=element;p->sym!=NULL;p++)
	{index=256*tolower(p->sym[1])+toupper(p->sym[0]);
	 wahr[index].a=p->w1; wahr[index].b=p->w2;
	 wahr[index].m1=p->m1; wahr[index].m2=p->m2;
	}
}

double *pas3eck(int n) //Pascalsches Dreieck
{
 int i,j;
 double *x=new double[(n+1)*(n+1)];
 for(i=0;i<n;i++)
  for(j=0;j<n;j++)
	if(i==0 || j==0) x[i+j*n]=1;
	else	x[i+j*n]=x[i-1+j*n]+x[i+(j-1)*n];
 return x;
}

struct NisoWahrtab {char sym[4]; double w[20];}; //Wahrscheinlichkeitstabelle fuer N Isotopenmarkierte
static NisoWahrtab isowahrtab[20];
static int isoanzahl=0;

void calcisowahr(const char *sym,int n)
{
 int i,j,index;
 double *feld;
 double wa,wb; //Wahrscheinlichkeiten fuer die 2 haeufigsten Isotope
 if(isoanzahl==20)
   {
     static int flag=0;
     if(flag==0) printf("Zu viele Atome. %s nicht mehr beruecksichtigt\n",sym);
     flag=1;
     return;
   }
 if(strlen(sym)>2) {printf("Fehler: ungueltiges Atom '%s'\n",sym); return;}
 strcpy(isowahrtab[isoanzahl].sym,sym);
 feld=pas3eck(n+1);
 if(argflag['V'])
  {printf("Wahrscheinlichkeitsberechnung fuer %d %s Atome\n",n,sym);//test
   for(j=n,i=0;j>=0 && i<20;i++,j--) printf(" %.0f",feld[i+j*(n+1)]); //test
   printf("\n");//test
  }
 index=256*tolower(sym[1])+toupper(sym[0]);
 wa=wahr[index].a; wb=wahr[index].b;
 for(j=n,i=0;j>=0 && i<20;i++,j--)
  {
   isowahrtab[isoanzahl].w[i] = pow(wa,j)*pow(wb,i)*feld[i+j*(n+1)];
   if(argflag['V']) printf(" %lf%%",100*isowahrtab[isoanzahl].w[i] ); //test
  }
 if(argflag['V']) printf("\n");//test
 isoanzahl++;
 delete feld;
}

/********************** Spektrometer-Simulation ***********************/
#define XSTEP 50
#define AUFLOESUNG (1.0/XSTEP)
#define BEREICH 20
#define IMAX (BEREICH*XSTEP) //Messbereich Multipliziert mit Anzahl aufgeloesten Punkten pro Masseneinheit
#define IMAX3 (3*IMAX)

class Spektrometer
{
 double molekuelmasse,startmasse,endmasse,startx,endx;
 uint detektorfeld[IMAX3];
public:
 Spektrometer() {startmasse=0.0; endmasse=BEREICH;}
 void init(const char *formel,int bereichsflag);
 void injekt(double masse);
 void drawxachse();
 void draw();
 void rahmenzeichnen();
};

static Spektrometer spektrum;

const double xmin= -1.0, ymin= -0.1, xmax=BEREICH+1.0, ymax=1.05;
const double ymark10= -0.06, ymark5= -0.05, ymark1= -0.04, ymark0= -0.03;
const double schrift_bx=xmax/80.0,schrift_hy=ymax/40.0;

void Spektrometer::init(const char *formel,int bereichsflag)
{
 int i,anzahl;
 char *sym;
 init_wahr();
 molekuelmasse = mg(formel);
 printf("Molmasse isotopenrein ist %f\n",molekuelmasse);
 startx = (int)molekuelmasse-1;
 endx = startx+BEREICH;
 startmasse = startx-BEREICH;
 endmasse = startx+2*BEREICH;
 if(bereichsflag!=0)
   {startx=startmasse; endx=endmasse;}
 printf("Messbereich: %f ... %f\n",startmasse,endmasse);//test
 //if(bereichsflag!=0)
  printf("Anzeigebereich: %f ... %f\n",startx,endx);//test
 for(i=0;i<IMAX3;i++) detektorfeld[i]=0;
 //printf("bx=%f hy=%f\n",schrift_bx,schrift_hy);//test
 textsize(schrift_bx,schrift_hy);

 //Aus der Formel fuer jede Atomsorte die Wahrscheinlichkeiten berechnen,
 //dass n Atome davon Isotope sein koennten:
 isoanzahl=0;
 while(*formel!=0)
  {
   sym=getsymbol(&formel);
   if(isdigit(*formel)) anzahl=getnumber(&formel);
   else anzahl=1;
   calcisowahr(sym,anzahl);
  }

 if(argflag['V']) printf("isoanzahl=%d\n",isoanzahl);//test
 //printf("sizeof(int)=%ld sizeof(ulong)=%ld\n",sizeof(int),sizeof(ulong));//test

}

void Spektrometer::injekt(double masse)
{
 double x,zuf,sumw;
 const double mgH=1.007825;
 int i,n,index;
 char isotop[40];
 masse += molekuelmasse + mgH; //Mit H+ Jonisiertes Molekuel

 if(masse>=startmasse && masse<=endmasse)
  for(i=0,x=startmasse;i<IMAX3;i++,x+=AUFLOESUNG)
   {
    if(x>=masse)
      {
       detektorfeld[i]++;
       break;
      }
   }
}

void Spektrometer::draw()
{
 int i,pen=PENUP;
 double x,y,max=100;
#ifdef SCHWARZAUFWEISS
 color(1);
#else
 color(0);
#endif
 fillbox(0.05,ymark0+0.005,BEREICH-0.05,ymax-0.005); //Bereich innerhalb Rahmen loeschen
#ifdef SCHWARZAUFWEISS
 color(0);
#else
 color(1);
#endif

 for(i=0;i<IMAX3;i++) if(detektorfeld[i]>max) max=detektorfeld[i];
 int i1=IMAX3/(endmasse-startmasse)*(startx-startmasse);
 int i2=IMAX3/(endmasse-startmasse)*(endx-startmasse);
 for(i=i1;i<i2;i++)
  {
   y = detektorfeld[i]/(double)max;
   x = BEREICH/(double)(i2-i1)*(i-i1);
   plot(x,y,pen);
   pen=PENDOWN;
  }
}

void Spektrometer::drawxachse()
{
 double xp,xw,y; //xp=Zeichnen-Korrdinaten, xw=Wirkliche x-Werte
 int ix;
 char text[40];
 //Zeichnen-Koordinaten immer 0.0 bis BEREICH (z.B. 20.0)
 plot(0.0,ymark0,PENUP); plot(BEREICH,ymark0,PENDOWN); //Achse zeichnen
 for(xw=startx;xw<=endx;xw+=1.0) //Markierstriche zeichnen
  {
   xp=BEREICH/(endx-startx)*(xw-startx);
   ix=xw;
   if(ix%5==0)
    {
     sprintf(text,"%.0f",xw);
     schrift(xp-strlen(text)*schrift_bx/2,ymark10-1.1*schrift_hy,text);
     y = (ix%10==0) ? ymark10 : ymark5;
    }
   else y=ymark1;
   plot(xp,y,PENUP); plot(xp,ymark0,PENDOWN);
  }
}

void Spektrometer::rahmenzeichnen()
{
 plot(0.0,ymark0,PENUP); plot(0.0,ymax,PENDOWN); plot(BEREICH,ymax,PENDOWN); plot(BEREICH,ymark0,PENDOWN);
}

void refresh()
{
 spektrum.draw();
 spektrum.rahmenzeichnen();
 spektrum.drawxachse();
}

/******************** Molekulargewicht-Berechnung *********************/
double mg(const char *formel)
{
 char str[200],*s;
 strcpy(str,formel);
 //error_reset(errortext,warningtext);
 return mol_weight(str,&s,ISO);
}

/************************* Menu Behandlung ****************************/
static int exitflag=0;
void menu_exit() {exitflag=1;}
void menu_refresh() {waitTOF(); inital_new(); refresh(); term_refresh();}

/************************* Hauptprogramm ******************************/
main(int argc,char *argv[])
{
 char formel[400];
 formel[0]=0;
 int i,j,c,col=0,maxcol,anzahlmessungen=1000;
 int breite,hoehe,tiefe,visklasse;
 //printf("Test: sizeof(uint)=%lu sizeof(ulong)=%lu sizeof(UINT64)=%lu\n",sizeof(uint),sizeof(ulong),sizeof(UINT64));
 //printf("Test: sizeof(float)=%lu sizeof(double)=%lu\n",sizeof(float),sizeof(double));
 if(argc<=0)
   ;/* es wurde von WorkBench gestartet */
 else
   /* es wurde von der Shell gestartet */
   for(j=0,i=1;i<argc;i++)
	{if((c= *argv[i])=='-' || c=='?') setargflags(argv[i]);
	 else	{if(++j==1) strcpy(formel,argv[i]);
                 else if(j==2) sscanf(argv[i],"%d",&anzahlmessungen);
	}	}
 if(argflag['?'] || j>MAXARG)
	{printf("mssimu  %s\n",VERSION);
	 printf("Anwendung: mssimu [-flags] Summenformel [AnzahlMessungen]\n");
	 printf("   Voreinstellung fuer AnzahlMessungen ist %d\n",anzahlmessungen);
	 printf("   Flags: v = verbose, einige Testausdrucke machen\n");
	 printf("          b = groesserer Bereich darstellen\n");
	 exit(0);
	}
 //tek_setdebug(1);//test

 if(*formel==0)
  {printf("Summenformel:"); do i=scanf("%s",formel); while(i!=1);}

 char *summenformel=getsummenformel(formel,NULL);
 if(strcmp(summenformel,formel)!=0)
   printf("Summenformel: %s\n",summenformel);

 getmaxsize(&breite,&hoehe,&tiefe,&visklasse);
 if(tiefe>TIEFE) tiefe=TIEFE;
 if(breite>XMAX) breite=XMAX;
 if(hoehe>YMAX) hoehe=YMAX;
 maxcol=(1<<TIEFE);
 setsize(breite,hoehe,tiefe);
 setmenu(1,"File");
 setmenu(1,"Refresh",&menu_refresh);
 setmenu(1,"Exit",&menu_exit);
 inital(xmin,ymin,xmax,ymax); /* Grafikfenster oeffnen */

#ifdef SCHWARZAUFWEISS
 //fuer schwarz auf weiss (default ist weiss auf schwarz):
 if(TIEFE==24)
  {screenclear(0xFFFFFF); rgbcolor(0x00,0x00,0x00);}
 else
  {screenclear(1); color(0);}
#endif

 int bereichsflag = argflag['B'] ? 1 : 0;
 spektrum.init(summenformel,bereichsflag);
 spektrum.rahmenzeichnen();
 spektrum.drawxachse();
 spektrum.draw(); //noch leeres Spektrum zeichnen
 term_refresh();
 inital_new();

 double multiplikator=1,summe;
 double proz[20];
 int jmax;
 printf("Es wird das Spektrum von %s berechnet.\n",summenformel);
 printf("(es sind bisher nur Formeln der Form CxHxOxSx erlaubt)\n");//test
 for(i=1;i<isoanzahl;i++)
   multiplikator *= isowahrtab[i].w[0];
 for(j=0;j<20;j++)
   proz[j] = multiplikator*isowahrtab[0].w[j];
 printf("Verteilung mit nur erstem Atom:\n");//test
 for(j=0;j<20 && proz[j]>=0.00005;j++) printf(" %.2f%%",proz[j]*100);//test
 jmax=j;
 printf("\n");//test
 //Vereinfachung: nur CxHxOxSx erlaubt
 if(isoanzahl!=5) printf("falsche Formel, %d Atomsorten (statt genau 5)\n",isoanzahl);//test
 for(j=1;j<jmax;j++)
  {
   summe=0;
   for(int k=1;k<=2;k++)
    {
     multiplikator = isowahrtab[0].w[0];
     for(i=1;i<isoanzahl;i++)
       multiplikator *= isowahrtab[i].w[(i==k)?j:0];
     summe += multiplikator;
    }
   if((j%2)==0)
    {
     for(int k=3;k<=4;k++)
      {
       multiplikator = isowahrtab[0].w[0];
       for(i=1;i<isoanzahl;i++)
	 multiplikator *= isowahrtab[i].w[(i==k)?(j/2):0];
       summe += multiplikator;
      }
    }
   proz[j] += summe;
  }
 printf("Gesamte Verteilung:\n");//test
 for(j=0;j<20 && proz[j]>=0.00005;j++) printf(" %.2f%%",proz[j]*100);//test
 printf("\n");//test

 for(j=0;j<jmax;j++)
  {
   for(i=0;i<anzahlmessungen*proz[j];i++)
    {
     spektrum.injekt(j);
     if(i<1000 || i%10000==0) {spektrum.draw(); waitTOF();}//test
    }
  }

 refresh();
 printf("fertig.\n");

 term_refresh();
 while(exitflag==0 && waitmenu(1)==0)
	;// auf Benutzereingaben warten
 term_exit();
 free(summenformel);
 return 0;
}/* ende von main */
